/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  DeadFTP
 *  Copyright (C) 2000-2001 Brandon Lees
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdarg.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pthread.h>
#include <signal.h>
#include <errno.h>
#include <ctype.h>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>
#include <glade/glade.h>
#include <gnome-xml/tree.h>
#include <gnome-xml/parser.h>

#include "globals.h"
#include "main.h"
#include "ftp.h"
#include "socket.h"
#include "hostmanager.h"
#include "remote.h"
#include "preferences.h"
#include "fxp.h"
#include "queueui.h"

/* Local function */
static void ftp_parse(char *buffer, int len);
gint ftp_retry(gpointer w);
static void ftp_do_connect(void);
static void ftp_do_disconnect(void);
void ftp_disconnect_retry(void);
void *ftp_read_list(void *arg);
void *ftp_read_data(void *data);
void *ftp_write_data(void *data);
void print_trans_message(void);
static void ftp_read_list_line(char *line, char *extradir);
void ftp_read(gpointer data, gint source, GdkInputCondition condition);
void remote_dblclick(void);
static void command_status(const char *code, int *, int *, int *);
static void ftp_data_socket_callback(socket_ctx_t * sock, gpointer data);
static void ftp_list_socket_callback(socket_ctx_t * sock, gpointer data);
static void ftp_control_socket_callback(socket_ctx_t * sock, gpointer data);
static gboolean quirk_compare(quirk_t * quirk, const char *server_string);
static void ftp_send_data(const char *, gboolean);
static void ftp_start_data_callback(socket_ctx_t * sock, gpointer data);
static void ftp_active_accept_conn(socket_ctx_t * sock, gpointer data);
static gboolean ftp_check_resume(queue_item_t * q, unsigned long size);
static void ftp_send_next_command(void);
static void ftp_cleanup_command(void);
static gboolean check_need_list(void);

gboolean ftp_parse_init(rcode_t reply, const char *arg);
gboolean ftp_parse_user(rcode_t reply, const char *arg);
gboolean ftp_parse_pass(rcode_t reply, const char *arg);
gboolean ftp_parse_list_pasv_port(rcode_t reply, const char *arg);
gboolean ftp_parse_cwd(rcode_t reply, const char *arg);
gboolean ftp_parse_start_list(rcode_t reply, const char *arg);
gboolean ftp_parse_trans_cwd(rcode_t reply, const char *arg);
gboolean ftp_parse_trans_size(rcode_t reply, const char *arg);
gboolean ftp_parse_trans_rest(rcode_t reply, const char *arg);
gboolean ftp_parse_trans_pasv(rcode_t reply, const char *arg);
gboolean ftp_parse_trans_retr(rcode_t reply, const char *arg);
gboolean ftp_parse_trans_type(rcode_t reply, const char *arg);
gboolean ftp_parse_sent_cwd(rcode_t reply, const char *arg);
gboolean ftp_parse_idle(rcode_t reply, const char *arg);
gboolean ftp_parse_doing_list(rcode_t reply, const char *arg);
gboolean ftp_parse_mkdir(rcode_t reply, const char *arg);
gboolean ftp_parse_delete(rcode_t reply, const char *arg);
gboolean ftp_parse_rename_from(rcode_t reply, const char *arg);
gboolean ftp_parse_rename_to(rcode_t reply, const char *arg);
gboolean ftp_parse_abort(rcode_t reply, const char *arg);
gboolean ftp_parse_fxp_size(rcode_t reply, const char *arg);
gboolean ftp_parse_fxp_pasv(rcode_t reply, const char *arg);
gboolean ftp_parse_fxp_port(rcode_t reply, const char *arg);
gboolean ftp_parse_fxp_trans(rcode_t reply, const char *arg);
gboolean ftp_parse_rest_test(rcode_t reply, const char *arg);
gboolean ftp_parse_rest_done(rcode_t reply, const char *arg);
gboolean ftp_parse_chmod(rcode_t reply, const char *arg);
gboolean ftp_parse_d_mkdir(rcode_t reply, const char *arg);

ftp_t FTP;

extern GMemChunk *afile_chunk;
GStringChunk *afile_remote_string_chunk = NULL;
extern GList *remote_files;

#define	STRCHR_NL(x,y)		do { \
					x = strchr(y, '\r'); \
					if (!x) \
						x = strchr(y, '\n'); \
				} while (0)
#define	NULL_NL(x)		while (x && ((*x == '\n') || (*x == '\r'))) *x++ = '\0';

quirk_t server_quirks[] = {
  {"Serv-U FTP-Server", FTP_QUIRK_RECURSIVE_L},
  {"Microsoft FTP Service", FTP_QUIRK_MICROSOFT},
  {NULL, 0}
};

/* I like having each state parsed by it's own function because it avoids
 * the clutter of the large ftp_parse function that existed previously with
 * the neverending switch statement.  It also works better with the new command
 * structure.  */
parse_desc_t parse_functions[] = {
  /* 0 - FTP_STATE_CONNECTING */
  {NULL},
  /* 1 - FTP_STATE_INIT */
  {ftp_parse_init},
  /* 2  - FTP_STATE_SENT_USER */
  {ftp_parse_user},
  /* 3  - FTP_STATE_SENT_PASS */
  {ftp_parse_pass},
  /* 4  - FTP_STATE_NEED_LIST */
  {ftp_parse_list_pasv_port},
  /* 5  - FTP_STATE_CWD */
  {ftp_parse_cwd},
  /* 6  - FTP_STATE_START_LIST */
  {ftp_parse_start_list},
  /* 7  - FTP_STATE_TRANS_CWD */
  {ftp_parse_trans_cwd},
  /* 8  - FTP_STATE_TRANS_SIZE */
  {ftp_parse_trans_size},
  /* 9  - FTP_STATE_TRANS_REST */
  {ftp_parse_trans_rest},
  /* 10  - FTP_STATE_TRANS_PASV */
  {ftp_parse_trans_pasv},
  /* 11 - FTP_STATE_TRANS_RETR */
  {ftp_parse_trans_retr},
  /* 12 - FTP_STATE_TRANS_TYPE */
  {ftp_parse_trans_type},
  /* 13 - FTP_STATE_IDLE */
  {ftp_parse_idle},
  /* 14 - FTP_STATE_DOING_LIST */
  {ftp_parse_doing_list},
  /* 15 - FTP_STATE_MKDIR */
  {ftp_parse_mkdir},
  /* 16 - FTP_STATE_DELETE */
  {ftp_parse_delete},
  /* 17 - FTP_STATE_RENAME_FROM */
  {ftp_parse_rename_from},
  /* 18 - FTP_STATE_RENAME_TO */
  {ftp_parse_rename_to},
  /* 19 - FTP_STATE_ABORT */
  {ftp_parse_abort},
  /* 20 - FTP_STATE_FXP_SIZE */
  {ftp_parse_fxp_size},
  /* 21 - FTP_STATE_FXP_PASV */
  {ftp_parse_fxp_pasv},
  /* 22 - FTP_STATE_FXP_PORT */
  {ftp_parse_fxp_port},
  /* 23 - FTP_STATE_FXP_TRANS */
  {ftp_parse_fxp_trans},
  /* 24 - FTP_STATE_REST_TEST */
  {ftp_parse_rest_test},
  /* 25 - FTP_STATE_REST_DONE */
  {ftp_parse_rest_done},
  /* 26 - FTP_STATE_CHMOD */
  {ftp_parse_chmod},
  /* 27 - FTP_STATE_D_TRANS_MKD */
  {ftp_parse_d_mkdir},
};

void ftp_init()
{
  bzero((char *) &FTP, sizeof(FTP));
  FTP.retry_tag = -1;
  FTP.data_tag = -1;
  FTP.gtag = -1;
  FTP.state = FTP_STATE_DISCONNECTED;

  FTP.sock = socket_new();
  FTP.datasock = socket_new();
}

void ftp_send_command(int state, unsigned long flags, const char *stat_message,
		      const char *command, ...)
{
  va_list args;
  command_t *cmd;
  /* If there are no commands on the command stack then this command should be
   * send immediately */
  gboolean send = (FTP.commands == NULL);

  g_return_if_fail(command != NULL);

  cmd = g_new0(command_t, 1);
  cmd->flags = flags;
  cmd->state = state;
  cmd->status = g_strdup(stat_message);
  va_start(args, command);
  cmd->command = g_strdup_vprintf(command, args);
  va_end(args);

  FTP.commands = g_list_append(FTP.commands, cmd);

  if (send)
    ftp_send_next_command();
}

static void ftp_send_data(const char *data, gboolean hide)
{
  char buffer[REGULAR_BUFFER];

  g_snprintf(buffer, REGULAR_BUFFER, "%s\r\n", data);

  send(FTP.sock->sockfd, buffer, strlen(buffer), 0);
  if (hide) {
    char *ch = strchr(buffer, ' ');
    if (ch) {
      strcpy(ch + 1, "(hidden)\n");
    }
  }
  print_console(CONSOLE_LOCAL, "%s", buffer);

}

static void ftp_send_next_command()
{
  command_t *cmd;
  int id;

  if (!FTP.commands)
    return;

  cmd = (command_t *) FTP.commands->data;
  g_return_if_fail(!FTP_CMD_FLAG_IS_SET(cmd, FTP_CMD_FLAG_SENT));
  ftp_send_data(cmd->command, FTP_CMD_FLAG_IS_SET(cmd, FTP_CMD_FLAG_PASS));
  FTP.state = cmd->state;
  FTP_CMD_FLAG_SET(cmd, FTP_CMD_FLAG_SENT);

  id =
    gtk_statusbar_get_context_id(GTK_STATUSBAR(deadftp.status_state),
				 STATUS_CTX_STATE);
  gtk_statusbar_push(GTK_STATUSBAR(deadftp.status_state), id, cmd->status);
}

static void ftp_cleanup_command()
{
  command_t *cmd;
  int id;

  g_return_if_fail(FTP.commands != NULL);

  cmd = (command_t *) FTP.commands->data;
  if (cmd->command);
  g_free(cmd->command);
  if (cmd->status)
    g_free(cmd->status);

  FTP.commands = g_list_remove(FTP.commands, cmd);
  g_free(cmd);

  id =
    gtk_statusbar_get_context_id(GTK_STATUSBAR(deadftp.status_state),
				 STATUS_CTX_STATE);
  gtk_statusbar_pop(GTK_STATUSBAR(deadftp.status_state), id);
}

static void ftp_parse(char *buffer, int len)
{
  char *ch;
  rcode_t reply;
  gboolean cleanup;

  g_return_if_fail(buffer != NULL);
  g_return_if_fail(len > 0);

  if (!FTP.commands)
    g_warning("file %s: line %d (%s): there is no command in ftp_parse", __FILE__,
	      __LINE__, __PRETTY_FUNCTION__);

  if (IS_SET(FTP_FLAG_MULTI)) {
    if (strncmp(FTP.scode, buffer, 4) != 0) {
      if (atoi(FTP.scode) == 230) {
	FTP.readme = g_list_append(FTP.readme, g_strdup(buffer));
	SET_FLAG(FTP_FLAG_HAVE_README);
      }
      return;
    }

    if (IS_SET(FTP_FLAG_HAVE_README) && deadftp_prefs.login_message) {
      GladeXML *xml;
      GtkWidget *text;
      GList *r;
      char *c, b[REGULAR_BUFFER];

      if (deadftp.showReadme)
        gtk_widget_destroy(deadftp.showReadme);

      xml = glade_xml_new(deadftp.gladefile, "showReadme");
      if (!xml) {
        g_error(_("Couldn't load the interface!"));
      } else {
	deadftp.showReadme = glade_xml_get_widget(xml, "showReadme");
	glade_xml_signal_autoconnect(xml);
	text = glade_xml_get_widget(xml, "readme");
	gtk_object_unref(GTK_OBJECT(xml));

        for (r = FTP.readme; r; r = g_list_next(r)) {
	  c = (char *) r->data;
	  g_snprintf(b, REGULAR_BUFFER, "%s\n", c);
	  gtk_text_insert(GTK_TEXT(text), NULL, NULL, NULL, b, -1);
	  g_free(c);
        }
      }
      g_list_free(FTP.readme);
      CLEAR_FLAG(FTP_FLAG_HAVE_README);
      FTP.readme = NULL;
    } else {
      GList *r;
      for (r = FTP.readme; r; r = g_list_next(r)) {
	g_free(r->data);
      }
      g_list_free(FTP.readme);
      CLEAR_FLAG(FTP_FLAG_HAVE_README);
      FTP.readme = NULL;
    }

    CLEAR_FLAG(FTP_FLAG_MULTI);
  } else if (buffer[3] == '-') {
    strncpy(FTP.scode, buffer, 3);
    if (atoi(FTP.scode) == 230) {	/* Info */
      FTP.readme = g_list_append(FTP.readme, g_strdup(buffer + 4));
      SET_FLAG(FTP_FLAG_HAVE_README);
    }
    FTP.scode[3] = ' ';
    FTP.scode[4] = '\0';
    SET_FLAG(FTP_FLAG_MULTI);
    return;
  }

  ch = strchr(buffer, ' ');
  if (!ch)
    return;
  CLEAR_FLAG(FTP_FLAG_MULTI);
  if (ch)
    *ch++ = '\0';
  command_status(buffer, &reply[FTP_RCODE_REPLY], &reply[FTP_RCODE_REASON],
		 &reply[FTP_RCODE_DETAIL]);

  /* The parse function will return TRUE if it has finished 
   * processing and state should be changed to IDLE.  Otherwise
   * it returns FALSE.  If the function doesn't exists, the default
   * is TRUE */
  cleanup = TRUE;
  if (parse_functions[FTP.state].handler)
    cleanup = parse_functions[FTP.state].handler(reply, ch);

  if (!FTP.commands)
    g_warning("file %s: line %d (%s): there is no command in ftp_parse", __FILE__,
	      __LINE__, __PRETTY_FUNCTION__);

  /* The rest of this functions can only happen when the state is IDLE */
  if (!cleanup)
    return;

  FTP.state = FTP_STATE_IDLE;

  /* check if the list need to be refreshed from this command */
  if (check_need_list()) {
    /* Start a list */
    ftp_send_command(FTP_STATE_START_LIST, 0, _("Refreshing Directory List"), "PWD");
  }

  /* command must be cleaned up after checking for list since we need the current
   * command flags to decide if the current list is dirty */
  ftp_cleanup_command();

  if (FTP.state == FTP_STATE_IDLE)
    ftp_send_next_command();
}

static gboolean check_need_list()
{
  g_return_val_if_fail(FTP.commands != NULL, FALSE);

  /* If the command doesn't require a list we don't need to do one
   * so just return */
  if (!FTP_CMD_FLAG_IS_SET(FTP_COMMAND(), FTP_CMD_FLAG_LIST))
    return FALSE;

  /* If this command requires a list but the next command does not,
   * do the list, otherwise do all spooled commands that need a list
   * before refreshing the list */
  if (!FTP.commands->next)
    return TRUE;

  if (FTP_CMD_FLAG_IS_SET((command_t *) FTP.commands->next->data, FTP_CMD_FLAG_LIST))
    return FALSE;

  /* We need to do a list */
  return TRUE;
}

gboolean ftp_parse_init(rcode_t reply, const char *arg)
{
  quirk_t *quirk;
  switch (reply[FTP_RCODE_REPLY]) {
  case FTP_REPLY_OK:
    FTP_QUIRK_CLEAR();
    for (quirk = server_quirks; quirk->server_string; quirk++) {
      if (quirk_compare(quirk, arg)) {
	FTP_QUIRK_ADD(quirk->quirks);
	g_warning("Added quirk %lx\n", quirk->quirks);
      }
    }
    ftp_send_command(FTP_STATE_SENT_USER, 0, _("Logging In"), "USER %s",
		     FTP.server.username);
    return TRUE;
  case FTP_REPLY_BAD:
    ftp_disconnect();
    return FALSE;
  default:
    ftp_disconnect_retry();
    return FALSE;
  }
  /* this should never be reached since the default case in the switch
   * returns but the compiler might complain without it.  */
  return FALSE;
}

gboolean ftp_parse_user(rcode_t reply, const char *arg)
{
  switch (reply[FTP_RCODE_REPLY]) {
  case FTP_REPLY_OK_MORE:
    ftp_send_command(FTP_STATE_SENT_PASS, FTP_CMD_FLAG_PASS, _("Logging In"),
		     "PASS %s", FTP.server.password);
    return TRUE;
  case FTP_REPLY_OK:
    if (IS_SET(FTP_FLAG_CHANGE_DIR)) {
      ftp_send_command(FTP_STATE_CWD, FTP_CMD_FLAG_LIST, _("Changing Directory"),
		       "CWD %s", FTP.remote_dir);
      CLEAR_FLAG(FTP_FLAG_CHANGE_DIR);
      if (IS_SET(FTP_FLAG_START_TRANS)) {
	SET_FLAG(FTP_FLAG_TRANS_CWD);
      }
    } else {
      /* set flag to cause file list after this function has completed */
      FTP_CMD_FLAG_SET(FTP_COMMAND(), FTP_CMD_FLAG_LIST);
    }
    return TRUE;
  case FTP_REPLY_BAD:
    ftp_disconnect();
    return FALSE;
  default:
    ftp_disconnect_retry();
    return FALSE;
  }
  return FALSE;
}

gboolean ftp_parse_pass(rcode_t reply, const char *arg)
{
  switch (reply[FTP_RCODE_REPLY]) {
  case FTP_REPLY_OK:
    ftp_send_command(FTP_STATE_REST_TEST, 0, _("Testing for Resume Support"),
		     "REST 1234");
    return TRUE;
    /* If reply == FTP_REPLY_BAD then we will never be able to (should never be able to) connect with 
     * the current l&p. some ftp servers incorrectly return a BAD reply instead of a BAD_RETRY reply */
  case FTP_REPLY_BAD:
    ftp_disconnect();
    return FALSE;
  default:
    ftp_disconnect_retry();
    return FALSE;
  }
  return FALSE;
}

gboolean ftp_parse_rest_test(rcode_t reply, const char *arg)
{
  switch (reply[FTP_RCODE_REPLY]) {
  case FTP_REPLY_OK:
  case FTP_REPLY_OK_MORE:
    /* The server supports resuming, or at least seems to */
    SET_FLAG(FTP_FLAG_REST_SUPPORT);
    /* Fallthrough */
  default:
    ftp_send_command(FTP_STATE_REST_DONE, 0, _("Testing for Resume Support"),
		     "REST 0");
    return TRUE;
  }
  return TRUE;
}

gboolean ftp_parse_rest_done(rcode_t reply, const char *arg)
{
  if (IS_SET(FTP_FLAG_CHANGE_DIR)) {
    ftp_send_command(FTP_STATE_CWD, FTP_CMD_FLAG_LIST, _("Changing Cirectory"),
		     "CWD %s", FTP.remote_dir);
    CLEAR_FLAG(FTP_FLAG_CHANGE_DIR);
    if (IS_SET(FTP_FLAG_START_TRANS)) {
      SET_FLAG(FTP_FLAG_TRANS_CWD);
    }
  } else {
    /* cause list after current command is finished */
    FTP_CMD_FLAG_SET(FTP_COMMAND(), FTP_CMD_FLAG_LIST);
  }
  return TRUE;
}

gboolean ftp_parse_cwd(rcode_t reply, const char *arg)
{
  return TRUE;
}

gboolean ftp_parse_start_list(rcode_t reply, const char *arg)
{
  char *line, *ch, *p;
  switch (reply[FTP_RCODE_REPLY]) {
  case FTP_REPLY_OK:
    line = g_strdup(arg);
    p = strchr(line, '"'), p++;
    ch = strchr(p, '"'), *ch++ = '\0';
    strncpy(FTP.server.current_directory, p, REGULAR_BUFFER);
    /* update the combo box that shows the current dir */
    remote_change_combo(FTP.server.current_directory);
    ftp_start_data_connection(FTP_STATE_NEED_LIST);
    g_free(line);
    break;
  default:
    break;
  }
  return TRUE;
}

gboolean ftp_parse_list_pasv_port(rcode_t reply, const char *arg)
{
  char *line, *ch, *p;
  switch (reply[FTP_RCODE_REPLY]) {
  case FTP_REPLY_OK:
    if (IS_SET(FTP_FLAG_PASV)) {
      line = g_strdup(arg);
      p = strchr(line, '('), p++;
      ch = strchr(p, ',');
      *ch++ = '\0';
      FTP.dataip = 0;
      FTP.dataip = (atoi(p) << 24) & 0xff000000;
      p = strchr(ch, ','), *p++ = '\0';
      FTP.dataip |= (atoi(ch) << 16) & 0x00ff0000;
      ch = strchr(p, ','), *ch++ = '\0';
      FTP.dataip |= (atoi(p) << 8) & 0x0000ff00;
      p = strchr(ch, ','), *p++ = '\0';
      FTP.dataip |= (atoi(ch) & 0xff);
      ch = strchr(p, ','), *ch++ = '\0';
      FTP.dataport = (atoi(p) << 8) & 0xff00;
      p = strchr(ch, ')'), *p++ = '\0';
      FTP.dataport |= (atoi(ch) & 0xff);

      socket_addr_connect(FTP.datasock, FTP.dataip, FTP.dataport,
			  ftp_list_socket_callback, NULL);
      g_free(line);
    } else {
      ftp_list_socket_callback(FTP.datasock, NULL);
    }
    break;
  default:
    /* Error */
    break;
  }
  return TRUE;
}

gboolean ftp_parse_doing_list(rcode_t reply, const char *arg)
{
  switch (reply[FTP_RCODE_REPLY]) {
  case FTP_REPLY_OK:
    FTP.keepalive = time(NULL);
    SET_FLAG(FTP_FLAG_TRANSFER_END);
    /* fallthrough */
  default:
    break;
  }
  return FALSE;
}

gboolean ftp_parse_trans_cwd(rcode_t reply, const char *arg)
{
  switch (reply[FTP_RCODE_REPLY]) {
  case FTP_REPLY_OK:
    ftp_send_command(FTP_STATE_TRANS_TYPE, 0, _("Starting Transfer"), "TYPE I");
    break;
  default:
    queue_error(_("Bad Directory"));	/* Print an error message to the console and move on to
					 * the next queue item */
    break;
  }
  return TRUE;
}

gboolean ftp_parse_trans_type(rcode_t reply, const char *arg)
{
  int state = 0;
  switch (reply[FTP_RCODE_REPLY]) {
  case FTP_REPLY_OK:
    if (QUEUE_CURRENT(FTP.queue)->fxp)
      state = FTP_STATE_FXP_SIZE;
    else
      state = FTP_STATE_TRANS_SIZE;

    ftp_send_command(state, 0, _("Determining File Size"), "SIZE %s",
		     QUEUE_CURRENT(FTP.queue)->remote_file);
    break;
  default:
    queue_error(_("Couldn't set transfer type"));	/* This probably isn't a fatal error, but... */
    break;
  }
  return TRUE;
}

gboolean ftp_parse_trans_size(rcode_t reply, const char *arg)
{
  unsigned long len;
  gboolean ok = TRUE;
  /* Update the file size if not sending so it can be used to determine resume action */
  if (reply[FTP_RCODE_REPLY] == FTP_REPLY_OK) {
    if (!QUEUE_GET_FLAG(QUEUE_CURRENT(FTP.queue), QUEUE_FLAG_SEND)) {
      QUEUE_CURRENT(FTP.queue)->size = atoi(arg);
      queue_update_item_size(QUEUE_CURRENT(FTP.queue));
    }
    queue_calculate_size();
  }
  /* We now have the file size as close to the actual size as we're going to get
   * so check for resuming here */
  len = atol(arg);

  if (QUEUE_GET_FLAG(QUEUE_CURRENT(FTP.queue), QUEUE_FLAG_SEND)) {
    /* When we're sending, we only have the length given by the SIZE */
    if (!len) {
      /* If SIZE didn't return a valid reply, transfer but don't resume */
      QUEUE_CURRENT(FTP.queue)->resume = 0;
      ok = TRUE;
    } else {
      ok = ftp_check_resume(QUEUE_CURRENT(FTP.queue), len);
    }
  } else {
    len = len ? len : QUEUE_CURRENT(FTP.queue)->size;
    if (!len) {
      /* If neither the size from SIZE nor the size from the LIST is valid
       * transfer the file without resuming */
      QUEUE_CURRENT(FTP.queue)->resume = 0;
      ok = TRUE;
    } else {
      ok = ftp_check_resume(QUEUE_CURRENT(FTP.queue), len);
    }
  }
  /* ok will be false when
   *  - The resume action was Ask and we're waiting for user input
   *  - The resume action was Skip
   */
  if (!ok) {
    /* If the resume action was skip then the queue has been moved on to
     * the next item and the status has been changed so we don't want to 
     * reset the status to IDLE */
    if (FTP.state == FTP_STATE_TRANS_SIZE)
      FTP.state = FTP_STATE_IDLE;
    return TRUE;
  }

  /* Continue with the appropriate action to start the transfer or abort on error */
  switch (reply[FTP_RCODE_REPLY]) {
  case FTP_REPLY_OK:
    ftp_start_data_connection(FTP_STATE_TRANS_PASV);
    break;
  case FTP_REPLY_BAD:
    if (reply[FTP_RCODE_REASON] == 0) {
      ftp_start_data_connection(FTP_STATE_TRANS_PASV);
    } else if (reply[FTP_RCODE_REASON] == 5) {
      /* The file doesn't exist.  If we are sending, ignore, otherwise it's an error */
      if (QUEUE_GET_FLAG(QUEUE_CURRENT(FTP.queue), QUEUE_FLAG_SEND)) {
	ftp_start_data_connection(FTP_STATE_TRANS_PASV);
	QUEUE_CURRENT(FTP.queue)->resume = 0;
      } else {
	queue_error(_("Remote file does not exist"));
      }
    } else {
      queue_error(_("Couldn't determine file size"));
    }
    break;
  default:
    break;
  }
  return TRUE;
}

gboolean ftp_parse_trans_pasv(rcode_t reply, const char *arg)
{
  char *line, *ch, *p;
  switch (reply[FTP_RCODE_REPLY]) {
  case FTP_REPLY_OK:
    // extract IP address and port from reply
    if (IS_SET(FTP_FLAG_PASV)) {
      line = g_strdup(arg);
      p = strchr(line, '('), p++;
      ch = strchr(p, ',');
      *ch++ = '\0';
      FTP.dataip = 0;
      FTP.dataip = (atoi(p) << 24) & 0xff000000;
      p = strchr(ch, ','), *p++ = '\0';
      FTP.dataip |= (atoi(ch) << 16) & 0x00ff0000;
      ch = strchr(p, ','), *ch++ = '\0';
      FTP.dataip |= (atoi(p) << 8) & 0x0000ff00;
      p = strchr(ch, ','), *p++ = '\0';
      FTP.dataip |= (atoi(ch) & 0xff);
      ch = strchr(p, ','), *ch++ = '\0';
      FTP.dataport = (atoi(p) << 8) & 0xff00;
      p = strchr(ch, ')'), *p++ = '\0';
      FTP.dataport |= (atoi(ch) & 0xff);

      socket_addr_connect(FTP.datasock, FTP.dataip, FTP.dataport,
			  ftp_data_socket_callback, NULL);
      g_free(line);
    } else {
      ftp_data_socket_callback(FTP.datasock, NULL);
    }
    break;
  default:
    /* Error */
    queue_error(_("Couldn't setup data connection"));
    break;
  }
  return TRUE;
}

gboolean ftp_parse_trans_rest(rcode_t reply, const char *arg)
{
  switch (reply[FTP_RCODE_REPLY]) {
  case FTP_REPLY_BAD:
    print_console(CONSOLE_ERROR,
		  _("Couldn't resume transfer, starting from the beginning"));
    QUEUE_CURRENT(FTP.queue)->resume = 0;
    /* fallthrough */
  case FTP_REPLY_OK:
  case FTP_REPLY_OK_MORE:
    if (QUEUE_GET_FLAG(QUEUE_CURRENT(FTP.queue), QUEUE_FLAG_SEND))
      ftp_send_command(FTP_STATE_TRANS_RETR, 0, _("Transferring File"), "STOR %s",
		       QUEUE_CURRENT(FTP.queue)->remote_file);
    else
      ftp_send_command(FTP_STATE_TRANS_RETR, 0, _("Transferring File"), "RETR %s",
		       QUEUE_CURRENT(FTP.queue)->remote_file);
    QUEUE_CURRENT(FTP.queue)->trans_start = time(0);
    return TRUE;
  default:
    break;
  }
  return TRUE;
}

gboolean ftp_parse_trans_retr(rcode_t reply, const char *arg)
{
  switch (reply[FTP_RCODE_REPLY]) {
  case FTP_REPLY_OK:
    FTP.keepalive = time(NULL);
    SET_FLAG(FTP_FLAG_TRANSFER_END);
    /* Fall Through */
  case FTP_REPLY_OK_MORE:
  case FTP_REPLY_OK_WAIT:
    return FALSE;
  default:
    queue_error(_("Couldn't retrieve file"));
    break;
  }
  return TRUE;
}

gboolean ftp_parse_mkdir(rcode_t reply, const char *arg)
{
  switch (reply[FTP_RCODE_REPLY]) {
  case FTP_REPLY_OK:
    /* OK */
    /* nothing to do */
    break;
  default:
    /* Error */
    /* TODO: message box ?? */
    break;
  }
  return TRUE;
}

gboolean ftp_parse_delete(rcode_t reply, const char *arg)
{
  GtkWidget *mbox;
  switch (reply[FTP_RCODE_REPLY]) {
  case FTP_REPLY_OK:
    /* went ok, refresh the directory, which will be done on return */
    break;
  default:
    /* Error */
    /* The error message (on my server at least) seems to be in the format
     * filename: error message.  Need to check other servers and look into
     * using this test in the message box */
    mbox =
      gnome_message_box_new(_("Couldn't Delete File"), GNOME_MESSAGE_BOX_ERROR,
			    GNOME_STOCK_BUTTON_OK, NULL);
    gtk_widget_show(mbox);
    break;
  }
  return TRUE;
}

gboolean ftp_parse_rename_from(rcode_t reply, const char *arg)
{
  GtkWidget *mbox;
  switch (reply[FTP_RCODE_REPLY]) {
  case FTP_REPLY_OK_MORE:
    /* FIXME: add a command extra info pointer or something */
    g_return_val_if_fail(FTP.rename != NULL, TRUE);

    ftp_send_command(FTP_STATE_RENAME_TO, FTP_CMD_FLAG_LIST, _("Renaming File"),
		     "RNTO %s", FTP.rename);
    break;
  default:
    /* Error */
    mbox =
      gnome_message_box_new(_("Couldn't Rename File"), GNOME_MESSAGE_BOX_ERROR,
			    GNOME_STOCK_BUTTON_OK, NULL);
    gtk_widget_show(mbox);
    break;
  }
  return TRUE;
}

gboolean ftp_parse_rename_to(rcode_t reply, const char *arg)
{
  GtkWidget *mbox;
  switch (reply[FTP_RCODE_REPLY]) {
  case FTP_REPLY_OK:
    /* OK, nothing to do, list will be done when we return true */
    break;
  default:
    mbox =
      gnome_message_box_new(_("Couldn't Rename File"), GNOME_MESSAGE_BOX_ERROR,
			    GNOME_STOCK_BUTTON_OK, NULL);
    gtk_widget_show(mbox);
  }
  g_free(FTP.rename);
  FTP.rename = NULL;
  return TRUE;
}

gboolean ftp_parse_abort(rcode_t reply, const char *arg)
{
  switch (reply[FTP_RCODE_REPLY]) {
  case FTP_REPLY_OK:
    /* If we got a reply from the ABOR command then any extra replies have been passed
     * through*/
    return TRUE;
  default:
    break;
  }
  return FALSE;
}

gboolean ftp_parse_idle(rcode_t reply, const char *arg)
{
  return TRUE;
}

gboolean ftp_parse_fxp_size(rcode_t reply, const char *arg)
{
  if (!QUEUE_CURRENT(FTP.queue)->fxp->slave) {
    /* master send the PASV */
    /* TODO: add an option to enable slave to be PASV */
    ftp_send_command(FTP_STATE_FXP_PASV, 0, _("Initiating FXP Transfer"), "PASV");
  } else {
    unsigned long addr;
    unsigned short port;
    pthread_mutex_lock(&QUEUE_CURRENT(FTP.queue)->fxp->data->lock);
    addr = QUEUE_CURRENT(FTP.queue)->fxp->data->address;
    port = QUEUE_CURRENT(FTP.queue)->fxp->data->port;
    pthread_mutex_unlock(&QUEUE_CURRENT(FTP.queue)->fxp->data->lock);
    /* slave should only get here after receicing port info from master */
    ftp_send_command(FTP_STATE_FXP_PORT, 0, _("Initiating FXP Transfer"),
		     "PORT %d,%d,%d,%d,%d,%d", (addr >> 24) & 0xff,
		     (addr >> 16) & 0xff, (addr >> 8) & 0xff, addr & 0xff,
		     (port >> 8) & 0xff, port & 0xff);
  }
  return TRUE;
}

gboolean ftp_parse_fxp_pasv(rcode_t reply, const char *arg)
{
  char *line, *ch, *p;
  switch (reply[FTP_RCODE_REPLY]) {
  case FTP_REPLY_OK:
    line = g_strdup(arg);
    p = strchr(line, '('), p++;
    ch = strchr(p, ',');
    *ch++ = '\0';
    /* Lock the lock on the shm */
    pthread_mutex_lock(&QUEUE_CURRENT(FTP.queue)->fxp->data->lock);
    QUEUE_CURRENT(FTP.queue)->fxp->data->address = (atoi(p) << 24) & 0xff000000;
    p = strchr(ch, ','), *p++ = '\0';
    QUEUE_CURRENT(FTP.queue)->fxp->data->address |= (atoi(ch) << 16) & 0x00ff0000;
    ch = strchr(p, ','), *ch++ = '\0';
    QUEUE_CURRENT(FTP.queue)->fxp->data->address |= (atoi(p) << 8) & 0x0000ff00;
    p = strchr(ch, ','), *p++ = '\0';
    QUEUE_CURRENT(FTP.queue)->fxp->data->address |= (atoi(ch) & 0xff);
    ch = strchr(p, ','), *ch++ = '\0';
    QUEUE_CURRENT(FTP.queue)->fxp->data->port = (atoi(p) << 8) & 0xff00;
    p = strchr(ch, ')'), *p++ = '\0';
    QUEUE_CURRENT(FTP.queue)->fxp->data->port |= (atoi(ch) & 0xff);
    pthread_mutex_unlock(&QUEUE_CURRENT(FTP.queue)->fxp->data->lock);

    fxp_set_status(QUEUE_CURRENT(FTP.queue)->fxp, FXP_STATUS_PORT);
    g_free(line);
    return FALSE;
  default:
    break;
  }
  return TRUE;
}

gboolean ftp_parse_fxp_port(rcode_t reply, const char *arg)
{
  switch (reply[FTP_RCODE_REPLY]) {
  case FTP_REPLY_OK:
    /* this is on the slave side */
    ftp_send_command(FTP_STATE_FXP_TRANS, 0, _("Transferring File"), "STOR %s",
		     QUEUE_CURRENT(FTP.queue)->remote_file);
    fxp_set_status(QUEUE_CURRENT(FTP.queue)->fxp, FXP_STATUS_GO);
    return FALSE;
  default:
    fxp_set_status(QUEUE_CURRENT(FTP.queue)->fxp, FXP_STATUS_STOP);
  }
  return TRUE;
}

gboolean ftp_parse_fxp_trans(rcode_t reply, const char *arg)
{
  switch (reply[FTP_RCODE_REPLY]) {
  case FTP_REPLY_OK:
    /* if we are the slave of an FXP and the next queue item is another
     * FXP where we are again the slave then wait for the master to
     * init the transfer */
    if (queue_check_fxp_go()) {
      queue_start();
      return FALSE;
    }
    break;
  default:
    break;
  }
  return TRUE;
}

gboolean ftp_parse_chmod(rcode_t reply, const char *arg)
{
  switch (reply[FTP_RCODE_REPLY]) {
  case FTP_REPLY_OK:
    /* list will be done on return */
    break;
  default:
    /* Error */
    /* TODO: error dialog/? */
    break;
  }
  return TRUE;
}

gboolean ftp_parse_d_mkdir(rcode_t reply, const char *arg)
{
  queue_item_t *q;
  switch (reply[FTP_RCODE_REPLY]) {
  case FTP_REPLY_OK:
    /* OK */
    g_return_val_if_fail(ftp_queue_current != NULL, TRUE);
    q = (queue_item_t *) ftp_queue_current->data;
    ftp_send_command(FTP_STATE_CWD, FTP_CMD_FLAG_LIST,
		     _("Entering destination directory"), "CWD %s/%s", q->remote_dir,
		     q->remote_file);
    break;
  default:
    /* Error */
    queue_error(_("Couldn't create destination directory."));
    break;
  }
  return TRUE;
}

static void command_status(const char *code, int *reply, int *reason, int *degree)
{
  if (reply) {
    *reply = code[0] - '0';
  }
  if (reason) {
    *reason = code[1] - '0';
  }
  if (degree) {
    *degree = code[2] - '0';
  }
}

void *ftp_write_data(void *data)
{
  int fd = -1;
  char buffer[BIG_BUFFER];
  int len, sent;
  gboolean done = FALSE;

  g_snprintf(buffer, BIG_BUFFER, "%s/%s", QUEUE_CURRENT(FTP.queue)->local_dir,
	     QUEUE_CURRENT(FTP.queue)->local_file);
  fd = open(buffer, O_RDONLY);
  if (QUEUE_GET_FLAG(QUEUE_CURRENT(FTP.queue), QUEUE_FLAG_RESUME))
    lseek(fd, QUEUE_CURRENT(FTP.queue)->resume, SEEK_SET);

  socket_set_nonblock(FTP.datasock, FALSE);
  while (!done) {
    pthread_testcancel();
    if ((len = read(fd, buffer, BIG_BUFFER)) <= 0)
      done = TRUE;
    else {
      pthread_testcancel();
      /* We want to send the whole buffer we just read */
      sent = 0;
      while ((sent != len) && !done) {
	int slen;

	errno = 0;
	pthread_testcancel();
	slen = send(FTP.datasock->sockfd, buffer + sent, len - sent, 0);
	pthread_testcancel();
	if (slen <= 0)
	  done = TRUE;
	else
	  sent += slen;
      }
      QUEUE_ITEM_LOCK(QUEUE_CURRENT(FTP.queue));
      /* Critical Section */
      QUEUE_CURRENT(FTP.queue)->total_trans += sent;
      /* Critical Section */
      QUEUE_ITEM_UNLOCK(QUEUE_CURRENT(FTP.queue));
      FTP.timeout = time(NULL);
    }
  }
  SET_FLAG(FTP_FLAG_THREAD_DONE);
  close(fd);
  socket_disconnect(FTP.datasock);
  return NULL;
}

void *ftp_read_data(void *data)
{
  int fd = -1;
  char buffer[BIG_BUFFER];
  int len, wrote;
  gboolean done = FALSE;

  g_snprintf(buffer, BIG_BUFFER, "%s/%s", QUEUE_CURRENT(FTP.queue)->local_dir,
	     QUEUE_CURRENT(FTP.queue)->local_file);
  if (!QUEUE_GET_FLAG(QUEUE_CURRENT(FTP.queue), QUEUE_FLAG_RESUME)) {
    unlink(buffer);
  }
  fd = open(buffer, O_WRONLY | O_CREAT, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
  if (fd < 0) {
    print_console(CONSOLE_ERROR, _("Couldn't open local file %s/%s: %s\n"),
		  QUEUE_CURRENT(FTP.queue)->local_dir,
		  QUEUE_CURRENT(FTP.queue)->local_file, g_strerror(errno));
    done = TRUE;
  } else {
    if (QUEUE_CURRENT(FTP.queue)->mode) {
      fchmod(fd, QUEUE_CURRENT(FTP.queue)->mode);
    }
    if (QUEUE_GET_FLAG(QUEUE_CURRENT(FTP.queue), QUEUE_FLAG_RESUME))
      lseek(fd, QUEUE_CURRENT(FTP.queue)->resume, SEEK_SET);
  }

  socket_set_nonblock(FTP.datasock, FALSE);

  while (!done) {
    pthread_testcancel();
    if ((len = recv(FTP.datasock->sockfd, buffer, BIG_BUFFER, 0)) <= 0)
      done = TRUE;
    else {
      pthread_testcancel();

      FTP.timeout = time(NULL);
      /* When writing to disk it should be able to write the whole buffer in one shot but
       * we need to check cause it's not guaranteed, plus wee need to check for errors in case
       * of disk full, etc. */
      wrote = 0;
      while ((wrote != len) && !done) {
	int wlen;

	wlen = write(fd, buffer + wrote, len - wrote);
	pthread_testcancel();
	if (wlen < 0) {
	  print_console(CONSOLE_ERROR, _("Couldn't write data to disk: %s\n"),
			g_strerror(errno));
	  done = TRUE;
	} else {
	  wrote += wlen;
	}
      }
      QUEUE_ITEM_LOCK(QUEUE_CURRENT(FTP.queue));
      /* Critical Section */
      QUEUE_CURRENT(FTP.queue)->total_trans += wrote;
      /* End Critical Section */
      QUEUE_ITEM_UNLOCK(QUEUE_CURRENT(FTP.queue));
    }
  }

  SET_FLAG(FTP_FLAG_THREAD_DONE);
  close(fd);
  socket_disconnect(FTP.datasock);
  return NULL;
}

void *ftp_read_list(void *arg)
{
  char buffer[BIG_BUFFER], extradir[SMALL_BUFFER], *ch, *p, *c;
  int len, done = 0, start = 1;
  GList *l;
  struct afile *x;

  for (l = remote_files; l; l = g_list_next(l)) {
    x = (struct afile *) l->data;
    g_chunk_free(x, afile_chunk);
  }
  g_list_free(remote_files);
  remote_files = NULL;
  /* free the strings */
  if (afile_remote_string_chunk)
    g_string_chunk_free(afile_remote_string_chunk);

  afile_remote_string_chunk = g_string_chunk_new(1024);
  if (FTP.extralen)
    memcpy(buffer, FTP.extradata, FTP.extralen);
  *extradir = '\0';

  socket_set_nonblock(FTP.datasock, FALSE);

  while (!done) {
    pthread_testcancel();
    if (FTP.extralen)
      memcpy(buffer, FTP.extradata, FTP.extralen);

    pthread_testcancel();
    len =
      recv(FTP.datasock->sockfd, buffer + FTP.extralen, BIG_BUFFER - FTP.extralen,
	   0);
    pthread_testcancel();
    if (len <= 0) {
      g_warning("Got error on socket: %s\n", g_strerror(errno));
      socket_disconnect(FTP.datasock);
      done = 1;
      continue;
    }
    len += FTP.extralen;
    FTP.extralen = 0;
    buffer[len] = '\0';
    fprintf(stderr, "%s\n\n", buffer);
    p = buffer;

    /* We may have error messages from ls being spit out at us */
    /* Some sites don't send a total line so look for filename: error message */
    if ((ch = strchr(p, ':')) && (*(ch + 1) == ' ')
	&& !FTP_QUIRK(FTP_QUIRK_MICROSOFT)) {
      while (start && p && (strncmp(p, ".:", 2) != 0)
	     && (strncmp(p, "total", 5) != 0)) {
	STRCHR_NL(ch, p);
	NULL_NL(ch);
	p = ch;
      }
    }
    start = 0;

    if (!p)
      continue;

    /* Ignore .: when we do an ls -lR */
    if (strncmp(p, ".:", 2) == 0) {
      STRCHR_NL(ch, p);
      NULL_NL(ch);
      p = ch;
    }

    /* Ignore total line */
    if (strncmp(p, "total", 5) == 0) {
      STRCHR_NL(ch, p);
      NULL_NL(ch);
      p = ch;
    }

    if (!p)
      continue;
    STRCHR_NL(ch, p);
    if (!ch) {			/* Not full line */
      strncpy(FTP.extradata, p, SMALL_BUFFER);
      FTP.extralen = strlen(FTP.extradata);
      continue;
    }
    NULL_NL(ch);
    pthread_testcancel();
    while (p) {
      /* get subdir name */
      /* if (IS_SET(FTP_FLAG_NEW_LIST_DIR)) { */
      if (p[strlen(p) - 1] == ':') {
	CLEAR_FLAG(FTP_FLAG_NEW_LIST_DIR);
	/* May need to change this but for now we will try to continue even * 
	 * if we can't find the new dir */
	p[strlen(p) - 1] = '\0';
	strncpy(extradir, p, SMALL_BUFFER);
	STRCHR_NL(c, ch);
	NULL_NL(c);
	p = ch;
	ch = c;
      }
      /* Ignore total line */
      if (strncmp(p, "total", 5) == 0) {
	STRCHR_NL(c, ch);
	NULL_NL(c);
	p = ch;
	ch = c;
      }

      pthread_testcancel();
      ftp_read_list_line(p, extradir);
      pthread_testcancel();
      if (!ch) {
	break;
      }
      STRCHR_NL(c, ch);
      if (!c) {
	if (ch) {
	  strncpy(FTP.extradata, ch, SMALL_BUFFER);
	  FTP.extralen = strlen(FTP.extradata);
	}
	break;
      }
      NULL_NL(c);
      p = ch;
      ch = c;
    }
  }
  CLEAR_FLAG(FTP_FLAG_ADD_LIST_QUEUE);
  SET_FLAG(FTP_FLAG_THREAD_DONE);
  return NULL;
}

static char **list_split_line(char *line, char *d)
{
  char **list, *start, *cline;
  int i, len;
  GList *l = NULL, *k;

  cline = g_strdup(line);
  start = strtok(cline, d);
  while (start) {
    l = g_list_append(l, start);
    start = strtok(NULL, d);
  }
  len = g_list_length(l);
  list = g_new(char *, len + 1);
  list[len] = NULL;
  k = l;
  for (i = 0; i < len; i++) {
    list[i] = g_strdup((gchar *) k->data);
    k = g_list_next(k);
  }
  g_list_free(l);
  g_free(cline);
  return list;
}

static int is_date(char **line, int where, int len)
{
  char buffer[REGULAR_BUFFER], month[SMALL_BUFFER];
  int date, t1, t2, r;

  /* If there aren't enough fields then this isn;t the date */
  if ((where + 2) >= len)
    return 0;
  /* If month length != 3 then not a date */
  if (strlen(line[where]) != 3)
    return 0;
  date = atoi(line[where + 1]);
  /* day of the month out of bounds */
  if ((date <= 0) || (date > 31))
    return 0;
  g_snprintf(buffer, REGULAR_BUFFER, "%s %s %s", line[where], line[where + 1],
	     line[where + 2]);
  r = sscanf(buffer, "%s %d %d:%d", month, &date, &t1, &t2);
  if ((r != 3) && (r != 4)) {
    return 0;
  }
  return 1;
}

static int list_array_find_date(char **line, int len)
{
  int where;

  /* We know what the style of Microsoft's list is */
  if (FTP_QUIRK(FTP_QUIRK_MICROSOFT))
    return 0;

  /* Date will usually start at the 5th element */
  for (where = 5; where >= 0; where--) {
    if (is_date(line, where, len))
      return where;
  }
  for (where = 5; line[where] != NULL; where++) {
    if (is_date(line, where, len))
      return where;
  }
  return -1;
}

static char *list_join(char **list, int start, int num)
{
  int len, i;
  char *str, *p;

  len = 0;
  for (i = start; i < start + num; i++)
    len += strlen(list[i]) + 1;

  if (!len)
    return NULL;

  str = g_new(char, len);
  *str = '\0';
  for (i = start; i < (start + num - 1); i++) {
    strcat(str, list[i]);
    strcat(str, " ");
  }
  strcat(str, list[i]);
  p = g_string_chunk_insert_const(afile_remote_string_chunk, str);
  g_free(str);
  return p;
}

static void ftp_read_list_line(char *line, char *extradir)
{
  char rdir[REGULAR_BUFFER], **lline, *p, *ch;
  int num, date_start, i;
  struct afile *f;

  if (!line)
    return;
  else if (!*line) {
    /* empty line is probably (hopefully) from a ls -lR when starting a new * 
     * directory */
    SET_FLAG(FTP_FLAG_NEW_LIST_DIR);
    return;
  }

  f = g_chunk_new0(struct afile, afile_chunk);

  lline = list_split_line(line, " ");
  num = 0;
  while (lline[num])
    num++;

  if (num <= 0) {
    g_chunk_free(f, afile_chunk);
    return;
  }

  /*
   * mode       link    owner   group   size    month   date    time|year       filename        [-> link dest]
   * mode       link    uid     gid     size    month   date    time|year       filename        [-> link dest]
   * mode       link    owner   size    month   time|year       filename        [->link dest]
   * Microsoft FTP
   * 01-09-01  09:44AM                24576 blah.doc
   * 10-30-00  10:31AM       <DIR>          blah
   *
   * There's probably all kinds of other different, wierd combinations out there */

  /* not all microsoft servers have weird lists so we need to
   * figure out if this one does.  The first field of the weird
   * list is the date and is exactly 8 chars so if field 0 isn't 8
   * chars clear the quirk */
  if (!IS_SET(FTP_FLAG_FOUND_DATE) && FTP_QUIRK(FTP_QUIRK_MICROSOFT)
      && strlen(lline[0]) != 8)
    FTP.quirks &= ~FTP_QUIRK_MICROSOFT;

  if (!IS_SET(FTP_FLAG_FOUND_DATE) && !FTP_QUIRK(FTP_QUIRK_MICROSOFT)) {
    FTP.date_start = list_array_find_date(lline, num);
    SET_FLAG(FTP_FLAG_FOUND_DATE);
  }
  date_start = FTP.date_start;

  if (FTP_QUIRK(FTP_QUIRK_MICROSOFT)) {
    if (g_strcasecmp(lline[2], "<DIR>") == 0) {
      f->is_dir = 1;
      f->date = list_join(lline, 0, 2);
      /* CHECKME: this will probably work, needs to be confirmed though. */
      f->file = g_string_chunk_insert_const(afile_remote_string_chunk, line + 39);
    } else {
      f->is_dir = 0;
      f->length = g_string_chunk_insert_const(afile_remote_string_chunk, lline[2]);
      f->date = list_join(lline, 0, 2);
      /* CHECKME: this will probably work, needs to be confirmed though. */
      f->file = g_string_chunk_insert_const(afile_remote_string_chunk, line + 39);
    }
  } else {
    g_return_if_fail(date_start != -1);
    /* Mode */
    f->mode = g_string_chunk_insert_const(afile_remote_string_chunk, lline[0]);
    if (*f->mode == 'd')
      f->is_dir = 1;
    else
      f->is_dir = 0;

    /* owner (maybe) */
    if (date_start >= 4) {
      f->owner = g_string_chunk_insert_const(afile_remote_string_chunk, lline[2]);
    }

    /* group if it's there */
    if (date_start >= 5) {
      f->group = g_string_chunk_insert_const(afile_remote_string_chunk, lline[3]);
    }

    /* Size */
    f->length =
      g_string_chunk_insert_const(afile_remote_string_chunk, lline[date_start - 1]);

    /* Date */
    g_snprintf(rdir, REGULAR_BUFFER, "%s %.2s %.5s", lline[date_start],
	       lline[date_start + 1], lline[date_start + 2]);
    f->date = g_string_chunk_insert_const(afile_remote_string_chunk, rdir);

    /* We probably want to save this since if we get it the server doesn't support changing
     * symlinks to dirs and files when we ask for a LIST -L */
    i = date_start + 3;
    for (i = 0, p = line; i < date_start + 3; i++) {
      while (!isspace(*p))
	p++;
      while (isspace(*p))
	p++;
    }
    ch = strstr(p, " ->");
    if (ch)
      *ch++ = '\0';
    f->file = g_string_chunk_insert_const(afile_remote_string_chunk, p);
  }

  for (i = 0; i < num; i++)
    g_free(lline[i]);
  g_free(lline);

  if (IS_SET(FTP_FLAG_ADD_LIST_QUEUE)) {
    char cwd[SMALL_BUFFER];
    getcwd(cwd, SMALL_BUFFER);
    g_snprintf(rdir, REGULAR_BUFFER, "%s/%s", FTP.server.current_directory,
	       extradir);
    add_queue(&FTP.server, f, cwd, f->file, FALSE);
    g_chunk_free(f, afile_chunk);
  } else {
    remote_files =
      g_list_insert_sorted(remote_files, (gpointer) f,
			   dir_sort_funcs[deadftp_prefs.remote_sort_field].func);
  }
}
void ftp_read(gpointer data, gint source, GdkInputCondition condition)
{
  int len;
  char buffer[REGULAR_BUFFER], *ch = NULL, *p, *c;

  /* need to check to make sure the source if the control socket since
   * the handler might be called for other sockets connecting */
  if (source == FTP.sock->sockfd && condition & GDK_INPUT_READ) {
    if (FTP.extracmd_len) {
      strncpy(buffer, FTP.extracmd, SMALL_BUFFER);
    }
    len =
      recv(FTP.sock->sockfd, buffer + FTP.extracmd_len,
	   REGULAR_BUFFER - FTP.extracmd_len, 0);
    if (len <= 0) {
      ftp_disconnect_retry();
      return;
    }

    len += FTP.extracmd_len;

    *FTP.extracmd = '\0';
    FTP.extracmd_len = 0;

    buffer[len] = '\0';
    FTP.timeout = time(NULL);
    p = buffer;
    STRCHR_NL(c, buffer);
    for (; c; c = strchr(p, '\r')) {
      ch = p;
      p = c;
      if (!p)
	break;
      while ((*p == '\r') || (*p == '\n'))
	*p++ = '\0';

      print_console(CONSOLE_REMOTE, "%s\n", ch);

      ftp_parse(ch, strlen(ch));
    }
    if (p && *p) {
      strncpy(FTP.extracmd, p, SMALL_BUFFER);
      FTP.extracmd_len = strlen(FTP.extracmd);
    }
  }
}

void ftp_disconnect_retry()
{
  int delay = 0;

  ftp_do_disconnect();

  print_console(CONSOLE_INFO, _("Disconnected...Waiting for retry.\n"));

  if (FTP.host_entry) {
    delay = hm_get_pref_int("retry_delay");
  }
  if (!FTP.host_entry || (delay < 0)) {
    delay = deadftp_prefs.retry_delay;
  }

  FTP.retry_tag = gtk_timeout_add(delay * 1000, ftp_retry, (gpointer) 0);
}

void ftp_disconnect()
{
  print_console(CONSOLE_INFO, _("Disconnected.\n"));

  ftp_do_disconnect();

  if (FTP.data_thread != 0)
    stop_data_thread();
  CLEAR_FLAG(FTP_FLAG_START_TRANS);

  hm_show_ctree();


}

static void ftp_do_disconnect()
{
  if (FTP.state == FTP_STATE_DISCONNECTED) {
    g_assert(FTP.commands == NULL);
    return;
  }

  ftp_clear_commands();

  socket_disconnect(FTP.sock);

  if (FTP.gtag != -1) {
    gdk_input_remove(FTP.gtag);
    FTP.gtag = -1;
  }
  FTP.state = FTP_STATE_DISCONNECTED;
  FTP.flags = 0;

  /*
   * gtk_statusbar_pop(GTK_STATUSBAR(deadftp.status1), status1_id);
   * gtk_statusbar_push(GTK_STATUSBAR(deadftp.status1), status1_id, _("Not Connected"));
   */

}

void ftp_connect()
{
  FTP.retry = 0;

  hm_hide_ctree();

  ftp_do_connect();
}

gint ftp_retry(gpointer w)
{
  int retry_count = 0;

  if (FTP.host_entry) {
    retry_count = hm_get_pref_int("retry_count");
  }
  if (!FTP.host_entry || (retry_count < 0)) {
    retry_count = deadftp_prefs.retry_count;
  }

  if (FTP.retry_tag != -1) {
    gtk_timeout_remove(FTP.retry_tag);
    FTP.retry_tag = -1;
  }

  if (FTP.retry <= retry_count) {
    FTP.retry++;
    ftp_do_connect();
  }
  return FALSE;
}

static void ftp_do_connect()
{
  command_t *cmd;

  if (FTP.state != FTP_STATE_DISCONNECTED) {
    ftp_disconnect();
    hm_hide_ctree();
  }

  print_console(CONSOLE_INFO, _("Connecting to %s:%d\n"), FTP.server.hostname,
		FTP.server.port);

  socket_connect(FTP.sock, FTP.server.hostname, FTP.server.port,
		 ftp_control_socket_callback, (gpointer) 0);

  /* add a command to the list since we now always want to have a command */
  cmd = g_new0(command_t, 1);
  cmd->status = g_strdup(_("Connecting..."));
  cmd->state = FTP_STATE_INIT;
  /* at this point, there should be no other commands on the stack */
  g_assert(FTP.commands == NULL);
  FTP.commands = g_list_append(FTP.commands, cmd);

  /* We are in CONNECTING State until we get the socket callback
   * saying we're connected to the server */
  FTP.state = FTP_STATE_CONNECTING;

  remote_update_quick_connect();
}

void ftp_exit(int status)
{
  if (FTP.state != FTP_STATE_DISCONNECTED)
    ftp_disconnect();
  save_preferences();
  save_hostmanager();
  gtk_exit(status);
}

void ftp_abort()
{
  /* We only need to abort if we are in the middle of a transfer */
  if (FTP.state != FTP_STATE_TRANS_RETR)
    return;

  stop_data_thread();

  /* threads are detached now so we shouldn't need to join them */
  /* pthread_join(FTP.data_thread, &arg); */
  FTP.data_thread = 0;
  queue_pop_status();

  print_console(CONSOLE_INFO, _("Transfer Aborted\n"));

  if (QUEUE_CURRENT(FTP.queue)) {
    queue_abort();
    QUEUE_CURRENT(FTP.queue) = NULL;
  }
}

void ftp_start_data_connection(int state)
{
  int use_pasv = 1;

  if (FTP.host_entry) {
    use_pasv = hm_get_pref_int("use_pasv");
  }
  if (!FTP.host_entry || (use_pasv < 0)) {
    use_pasv = deadftp_prefs.firewall.use_pasv;
  }

  if (use_pasv) {
    ftp_send_command(state, 0, _("Initiating Transfer"), "PASV");
    SET_FLAG(FTP_FLAG_PASV);
  } else
    socket_new_server(FTP.datasock, 0, ftp_start_data_callback,
		      GINT_TO_POINTER(state));
}

static void ftp_start_data_callback(socket_ctx_t * sock, gpointer data)
{
  char *sock_str;
  int state = GPOINTER_TO_INT(data);

  if (!sock || FTP.datasock->sockfd < 0) {
    FTP.state = FTP_STATE_IDLE;
    print_console(CONSOLE_ERROR, _("Couldn't create server socket: %s\n"),
		  g_strerror(errno));
    return;
  }

  /* set right TOS */
  socket_set_data(FTP.datasock);

  sock_str = socket_make_string(FTP.datasock);
  if (sock_str) {
    ftp_send_command(state, 0, _("Initiating Transfer"), "PORT %s", sock_str);
    CLEAR_FLAG(FTP_FLAG_PASV);
    socket_accept_connection(FTP.datasock, ftp_active_accept_conn, (gpointer) 0);
  }
  g_free(sock_str);
}

static void ftp_active_accept_conn(socket_ctx_t * sock, gpointer data)
{
  if (!sock || sock->sockfd == -1) {
    /* Error */
    return;
  }
  socket_set_data(sock);

  if (FTP.state == FTP_STATE_DOING_LIST) {
    create_data_thread(LIST_THREAD, ftp_read_list, NULL);
  } else {
    if (QUEUE_GET_FLAG(QUEUE_CURRENT(FTP.queue), QUEUE_FLAG_SEND))
      create_data_thread(DATA_THREAD, ftp_write_data, NULL);
    else
      create_data_thread(DATA_THREAD, ftp_read_data, NULL);
  }
}

gboolean ftp_check_resume(queue_item_t * item, unsigned long length)
{
  struct stat sbuf;
  int condition = 0;
  resume_t action = 0;
  int source, dest;

  if (stat(item->local_file, &sbuf) == -1) {
    item->resume = 0;
    QUEUE_CLEAR_FLAG(item, QUEUE_FLAG_RESUME);
    return TRUE;
  }
  if (QUEUE_GET_FLAG(item, QUEUE_FLAG_SEND)) {
    /* if we're sending the source file is the local file and
     * the destination file is the remote */
    source = sbuf.st_size;
    dest = length;
  } else {
    /* when downloading the source file is the remote and 
     * dest file is local */
    source = length;
    dest = sbuf.st_size;
  }

  /* Check fro flags from user pressing resume all, skip all, etc */
  switch (FTP.queue.resume_action) {
  case RESUME_OVERWRITE:
    item->resume = 0;
    QUEUE_CLEAR_FLAG(item, QUEUE_FLAG_RESUME);
    return TRUE;
  case RESUME_RESUME:
    /* If resume all was pressed but this server doesn't support
     * resuming fall through to other checks */
    if (IS_SET(FTP_FLAG_REST_SUPPORT)) {
      item->resume = dest;
      QUEUE_SET_FLAG(item, QUEUE_FLAG_RESUME);
      return TRUE;
    }
    break;
  case RESUME_SKIP:
    queue_next(FALSE);
    return FALSE;
  default:
    break;
  }
  /* RESUME_ASK will fall through from the above switch */

  /* Destination file is bigger than the source file */
  if (dest > source)
    condition = RESUME_COND_BIGGER;
  /* destination file is equal size to local file */
  else if (dest == source)
    condition = RESUME_COND_EQUAL;
  /* destination file is smaller that source file */
  else if (dest < source)
    condition = RESUME_COND_SMALLER;

  /* Figure out the correct action to take */
  if (QUEUE_GET_FLAG(item, QUEUE_FLAG_SEND))
    action = deadftp_prefs.resume_upload[condition];
  else
    action = deadftp_prefs.resume_download[condition];

  /* TRUE is returned if the correct action was determined, FALSE
   * if we need to wait for user input */
  switch (action) {
  case RESUME_ASK:
    item->resume = dest;
    load_resume_dialog(item);
    return FALSE;
  case RESUME_RESUME:
    /* Only resume if the server has resume support */
    if (IS_SET(FTP_FLAG_REST_SUPPORT)) {
      item->resume = dest;
      QUEUE_SET_FLAG(item, QUEUE_FLAG_RESUME);
      return TRUE;
    }
    /* Else fallthrough to overwrite */
  case RESUME_OVERWRITE:
    item->resume = 0;
    QUEUE_CLEAR_FLAG(item, QUEUE_FLAG_RESUME);
    return TRUE;
  case RESUME_SKIP:
    /* skip this one */
    queue_next(FALSE);
    return FALSE;
  }

  /* Should never be reached */
  return TRUE;
}

void create_data_thread(thread_type type, void *(*function) (void *), void *arg)
{
  pthread_attr_t attr;

  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
  pthread_create(&FTP.data_thread, &attr, function, arg);

  CLEAR_FLAG(FTP_FLAG_TRANSFER_DONE);
  if (type == DATA_THREAD) {
    gtk_progress_configure(GTK_PROGRESS(deadftp.status_progress), 0, 0,
			   QUEUE_CURRENT(FTP.queue)->size);
    FTP.data_tag = gtk_timeout_add(1000, ftp_timeout_data, NULL);
  } else if (type == LIST_THREAD)
    FTP.data_tag = gtk_timeout_add(1000, ftp_timeout_list, NULL);
}

gint ftp_timeout_data(gpointer data)
{
  queue_update_status();

  // We need to check with == since FLAG_TRANSFER_DONE is a combination of two flags
  if (IS_SET(FTP_FLAG_TRANSFER_DONE)) {
    FTP.data_thread = 0;
    queue_pop_status();
    print_trans_message();
    FTP.state = FTP_STATE_IDLE;
    ftp_cleanup_command();
    FTP.keepalive = time(NULL);
#if 0
    if (!QUEUE_GET_FLAG(QUEUE_CURRENT(FTP.queue), QUEUE_FLAG_SEND)) {
      extern void change_local(void);
      change_local();
      queue_next();
    } else {
      ftp_start_data_connection(FTP_STATE_NEED_LIST);
      SET_FLAG(FTP_FLAG_START_TRANS);
    }
#endif
    CLEAR_FLAG(FTP_FLAG_TRANSFER_DONE);
    FTP.data_tag = -1;
    queue_next(TRUE);

    return FALSE;
  }
  return TRUE;
}

gint ftp_timeout_list(gpointer data)
{
  if (IS_SET(FTP_FLAG_TRANSFER_DONE)) {
    FTP.data_thread = 0;
    gtk_clist_clear(GTK_CLIST(deadftp.remote));
    g_list_free(remote_row);
    remote_row = NULL;

    /* unfreeze queue if we froze it for an add dir */
    if (IS_SET(FTP_FLAG_FREEZE_QUEUE)) {
      gtk_clist_thaw(GTK_CLIST(deadftp.queue));
      CLEAR_FLAG(FTP_FLAG_FREEZE_QUEUE);
    }

    if (remote_files)
      remote_update_dir_list();

    if (IS_SET(FTP_FLAG_CHANGE_DIR_BACK)) {
      ftp_send_command(FTP_STATE_CWD, 0, _("Chaning Directory"), "CWD %s",
		       FTP.remote_dir);
      CLEAR_FLAG(FTP_FLAG_CHANGE_DIR_BACK);
    }
    FTP.state = FTP_STATE_IDLE;
    /* Cleanup the LIST command */
    ftp_cleanup_command();

    if (IS_SET(FTP_FLAG_START_TRANS)) {
      CLEAR_FLAG(FTP_FLAG_START_TRANS);
      queue_start();
    }
    // If queue_start did not start another transfer send the next command
    if (FTP.state == FTP_STATE_IDLE)
      ftp_send_next_command();

    CLEAR_FLAG(FTP_FLAG_TRANSFER_DONE);
    FTP.data_tag = -1;
    return FALSE;
  }
  return TRUE;
}

void stop_data_thread()
{
  if (FTP.data_thread != 0)
    pthread_cancel(FTP.data_thread);
  if (FTP.data_tag != -1) {
    gtk_timeout_remove(FTP.data_tag);
    FTP.data_tag = -1;
  }
}

static void ftp_control_socket_callback(socket_ctx_t * sock, gpointer data)
{
  char buffer[SMALL_BUFFER];

  if (!sock || sock->sockfd < 0) {
    print_console(CONSOLE_ERROR, _("Error: couldn't connect to %s:%d\n"),
		  FTP.server.hostname, FTP.server.port);
    ftp_disconnect_retry();
    hm_show_ctree();
    return;
  } else {
    print_console(CONSOLE_INFO, _("Connected to %s:%d waiting for reply...\n"),
		  FTP.server.hostname, FTP.server.port);
  }
  FTP.connect = time(0);
  gtk_timeout_add(1000, remote_connect_time, (gpointer) 0);
  FTP.state = FTP_STATE_INIT;
  FTP.timeout = time(NULL);
  FTP.keepalive = time(NULL);
  /* Proxy connections often have data available before we can
   * add the input callback so we need to check for it */
  if (recv(sock->sockfd, buffer, SMALL_BUFFER, MSG_PEEK) > 0)
    ftp_read(NULL, sock->sockfd, GDK_INPUT_READ);
  /* After we have cleared out data, add input handler.
   * XXX: is there a possbile race condition here were data comes in after we
   * read it but before the callback is added?  Would it work to add the input
   * handler then check for data on the socket?  */
  FTP.gtag = gdk_input_add(sock->sockfd, GDK_INPUT_READ, ftp_read, (gpointer) 0);
}

static void ftp_list_socket_callback(socket_ctx_t * sock, gpointer data)
{
  int show_hidden = 0;
  char extraflags[32];

  if (!sock || sock->sockfd < 0) {
    g_warning("Got error on socket: %s\n", g_strerror(errno));
    /* Is there anything we need to do here?  The socket failed to connect. */
    return;
  }
  FTP.extralen = 0;
  if (IS_SET(FTP_FLAG_PASV))
    create_data_thread(LIST_THREAD, ftp_read_list, NULL);

  *extraflags = '\0';
  if (FTP.host_entry)
    show_hidden = hm_get_pref_int("show_hidden");

  if (!FTP.host_entry || (show_hidden < 0))
    show_hidden = deadftp_prefs.list_method;

  if (show_hidden == LIST_SHOW_HIDDEN)
    strncat(extraflags, "a", 32);

  if (IS_SET(FTP_FLAG_RECURSIVE_LIST)) {
    strncat(extraflags, "R", 32);
    CLEAR_FLAG(FTP_FLAG_RECURSIVE_LIST);
  }
  if (FTP_QUIRK(FTP_QUIRK_RECURSIVE_L))
    ftp_send_command(FTP_STATE_DOING_LIST, 0, _("Listing Files"), "LIST -l%s",
		     extraflags);
  else
    ftp_send_command(FTP_STATE_DOING_LIST, 0, _("Listing Files"), "LIST -L%s",
		     extraflags);
}

static void ftp_data_socket_callback(socket_ctx_t * sock, gpointer data)
{
  if (!sock || sock->sockfd < 0) {
    /* WE should get an error message on the control socket that will cleanup. */
    return;
  }

  /* set right TOS */
  socket_set_data(sock);

  if (IS_SET(FTP_FLAG_PASV)) {
    if (QUEUE_GET_FLAG(QUEUE_CURRENT(FTP.queue), QUEUE_FLAG_SEND))
      create_data_thread(DATA_THREAD, ftp_write_data, NULL);
    else
      create_data_thread(DATA_THREAD, ftp_read_data, NULL);
  }

  QUEUE_CURRENT(FTP.queue)->total_trans = 0;
  if (QUEUE_GET_FLAG(QUEUE_CURRENT(FTP.queue), QUEUE_FLAG_RESUME)
      && IS_SET(FTP_FLAG_REST_SUPPORT)) {
    ftp_send_command(FTP_STATE_TRANS_REST, 0, _("Resuming Transfer"), "REST %d",
		     QUEUE_CURRENT(FTP.queue)->resume);
  } else {
    if (QUEUE_GET_FLAG(QUEUE_CURRENT(FTP.queue), QUEUE_FLAG_SEND))
      ftp_send_command(FTP_STATE_TRANS_RETR, 0, _("Sending File"), "STOR %s",
		       QUEUE_CURRENT(FTP.queue)->remote_file);
    else
      ftp_send_command(FTP_STATE_TRANS_RETR, 0, _("Receiving File"), "RETR %s",
		       QUEUE_CURRENT(FTP.queue)->remote_file);

    QUEUE_CURRENT(FTP.queue)->trans_start = time(0);
  }
}

static gboolean quirk_compare(quirk_t * quirk, const char *server_string)
{
  char *ch, *line;

  // try to match the string at each word
  line = g_strdup(server_string);
  ch = line;
  while (ch) {
    if (g_strncasecmp(quirk->server_string, ch, strlen(quirk->server_string)) == 0) {
      g_free(line);
      return TRUE;
    }
    // start the compare at the next word
    ch = strchr(ch, ' ');
    if (ch)
      ch++;
  }
  g_free(line);
  return FALSE;
}

void ftp_clear_commands()
{
  GList *l;

  for (l = FTP.commands; l; l = l->next) {
    command_t *cmd;

    cmd = (command_t *) l->data;
    if (cmd->command)
      g_free(cmd->command);
    if (cmd->status)
      g_free(cmd->status);
    g_free(cmd);
  }
  g_list_free(FTP.commands);
  FTP.commands = NULL;
}
