/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  DeadFTP
 *  Copyright (C) 2000-2001 Brandon Lees
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdarg.h>
#include <dirent.h>
#include <unistd.h>
#include <sys/stat.h>
#include <errno.h>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>
#include <glade/glade.h>
#include <gnome-xml/tree.h>
#include <gnome-xml/parser.h>

#include "globals.h"
#include "main.h"
#include "ftp.h"
#include "fxp.h"
#include "queueui.h"
#include "view.h"
#include "clipboard.h"

/*
 * Local function 
 */
static queue_item_t *queue_find_start(void);
queue_item_t *queue_find_next(const GList *);
void queue_item_delete(queue_item_t * q);
static gboolean queue_expand_directory(const queue_item_t * q);
static time_t queue_parse_date(const char *date);

extern void change_local(void);
extern GList *remote_files;
extern GList *local_files;

GList *ftp_queue = NULL, *ftp_queue_current;
queueid_t queueid = 1;

void queue_start()
{
  queue_item_t *q;

  if (IS_SET(FTP_FLAG_DONT_TRANSFER)) {
    CLEAR_FLAG(FTP_FLAG_DONT_TRANSFER);
    return;
  }

  q = queue_find_next(ftp_queue);
  if (!q)
    return;

  queue_start_item(q);
}

void queue_start_item(queue_item_t * q)
{
  gboolean new_host;
  char cwd[REGULAR_BUFFER];

  QUEUE_CURRENT(FTP.queue) = q;
  QUEUE_SET_STATUS(q, QUEUE_STATUS_ACTIVE);
  queue_update_item_status(q);

  if (q->fxp && fxp_start_transfer(q)) {
    /* move current back so we can call gueue_go again to start
     * at the right place */
    ftp_queue_current = g_list_previous(ftp_queue_current);
    return;
  }

  new_host = 0;
  if (FTP.state == FTP_STATE_DISCONNECTED) {
    /*
     * If we're not connected to a host then we obviously need to connect 
     */
    new_host = 1;
  }
  if (g_strcasecmp(FTP.server.hostname, q->hostname) != 0) {
    new_host = 1;
    strncpy(FTP.server.hostname, q->hostname, SMALL_BUFFER);
  }
  if (FTP.server.port != q->port) {
    new_host = 1;
    FTP.server.port = q->port;
  }
  if (g_strcasecmp(FTP.server.username, q->username) != 0) {
    new_host = 1;
    strncpy(FTP.server.username, q->username, 32);
  }
  if ((g_strcasecmp(FTP.server.password, q->password) != 0)
      && (g_strcasecmp(FTP.server.username, "anonymous") != 0)) {
    new_host = 1;
    strncpy(FTP.server.password, q->password, SMALL_BUFFER);
  } else if (g_strcasecmp(FTP.server.username, "anonymous") == 0) {
    strncpy(FTP.server.password, deadftp_prefs.anonymous_password, SMALL_BUFFER);
  }
  getcwd(cwd, REGULAR_BUFFER);
  if (new_host) {
    /*
     * start_queue will be called after we reconnect so we want the current
     * entry to be next so it doesn't * get skipped 
     */
    ftp_queue_current = g_list_previous(ftp_queue_current);
    /* We only need to disconnect if we connected in the first place */
    if (FTP.state != FTP_STATE_DISCONNECTED) {
      ftp_disconnect();
    }
    SET_FLAG(FTP_FLAG_START_TRANS);
    FTP.host_entry = NULL;

    strncpy(FTP.remote_dir, q->remote_dir, SMALL_BUFFER);
    SET_FLAG(FTP_FLAG_CHANGE_DIR);

    ftp_connect();
    FTP.state = FTP_STATE_INIT;
    return;
  }

  if (QUEUE_GET_FLAG(q, QUEUE_FLAG_DIRECTORY)) {
    /* First create the destination directory, if we can't do that all is
     * lost and this queue item will be marked with an error */
    if (QUEUE_GET_FLAG(q, QUEUE_FLAG_SEND)) {
      g_snprintf(cwd, REGULAR_BUFFER, "%s/%s", q->remote_dir, q->remote_file);
      if (g_strcasecmp(FTP.server.current_directory, cwd) != 0
	  && !IS_SET(FTP_FLAG_TRANS_CWD)) {
	/* If we're not already in the right dir then it needs to be create */
	ftp_send_command(FTP_STATE_D_TRANS_MKD, 0, _("Making Destination Directory"),
			 "MKD %s/%s", q->remote_dir, q->remote_file);
	return;
      }
    } else {
      g_snprintf(cwd, REGULAR_BUFFER, "%s/%s", q->local_dir, q->local_file);
      if (((mkdir(cwd, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) < 0
	    && errno != EEXIST)) || (chdir(cwd) < 0)) {
	queue_error(_("Couldn't create destination Directory"));
	return;
      }
      change_local();
      g_snprintf(cwd, REGULAR_BUFFER, "%s/%s", q->remote_dir, q->remote_file);
    }
    /* If we are already in the right directory then we can just fill the
     * queue from the cached list and don't need to refresh it, otherwise
     * we need to CWD on the server to get the right directory list */
    /* This will also fall through from the send condition above if we're already
     * in the destination dir so we can just fill up the queue from the local list */
    if (g_strcasecmp(FTP.server.current_directory, cwd) == 0
	|| IS_SET(FTP_FLAG_TRANS_CWD)) {
      /* all new items will be added before ftp_queue_current.  By moving the pointer one
       * back in the list it is easy to find the item we want to start with by moving it 
       * forward again */
      ftp_queue_current = g_list_previous(ftp_queue_current);

      queue_expand_directory(q);
      gtk_clist_remove(GTK_CLIST(deadftp.queue), g_list_index(ftp_queue, q));
      ftp_queue_current = g_list_next(ftp_queue_current);
      ftp_queue = g_list_remove(ftp_queue, q);
      queue_item_delete(q);

      CLEAR_FLAG(FTP_FLAG_TRANS_CWD);
      /* The queue may be empty if the directory was empty */
      if (ftp_queue_current)
	ftp_queue_current = g_list_next(ftp_queue_current);
      else
	ftp_queue_current = ftp_queue;

      if (ftp_queue_current)
	queue_start_item((queue_item_t *) ftp_queue_current->data);
      /* In any case, we're done */
      return;
    } else {
      ftp_send_command(FTP_STATE_CWD, FTP_CMD_FLAG_LIST, _("Changing Directory"),
		       "CWD %s/%s", q->remote_dir, q->remote_file);
      g_snprintf(FTP.server.current_directory, REGULAR_BUFFER, "%s/%s",
		 q->remote_dir, q->remote_file);
      SET_FLAG(FTP_FLAG_START_TRANS);
      ftp_queue_current = g_list_previous(ftp_queue_current);
      SET_FLAG(FTP_FLAG_TRANS_CWD);
      return;
    }
  }
  g_snprintf(cwd, REGULAR_BUFFER, "%s/%s", q->local_dir, q->local_file);
  /* If FTP_FLAG_TRANS_CWD is set, we've already sent CWD.  This is to handle symlink dir
   * names in the queue path since FTP.cwd will be the full path with no symlinks */
  if ((g_strcasecmp(FTP.server.current_directory, q->remote_dir) != 0)
      && !IS_SET(FTP_FLAG_TRANS_CWD)) {
    ftp_send_command(FTP_STATE_CWD, FTP_CMD_FLAG_LIST, _("Changing Directory"),
		     "CWD %s", q->remote_dir);
    strncpy(FTP.server.current_directory, q->remote_dir, REGULAR_BUFFER);
    SET_FLAG(FTP_FLAG_START_TRANS);
    /* move current back so when the transfer is resumed after changing directories, we will
     * resume at this item */
    ftp_queue_current = g_list_previous(ftp_queue_current);
    SET_FLAG(FTP_FLAG_TRANS_CWD);
  } else {
    ftp_send_command(FTP_STATE_TRANS_TYPE, 0, _("Setting Transfer Type"), "TYPE I");
    CLEAR_FLAG(FTP_FLAG_TRANS_CWD);
  }
}

void queue_next(gboolean remove)
{
  queue_item_t *q;
  GList *next;

  /* This is called when we just finished one item and are going on to
   * the next so we should always have a current item....check this */
  g_return_if_fail(ftp_queue_current != NULL);
  g_return_if_fail(QUEUE_CURRENT(FTP.queue) != NULL);

  /* remove current item from queue */
  q = (queue_item_t *) ftp_queue_current->data;
  next = ftp_queue_current->next;
  if (remove) {
    gtk_clist_remove(GTK_CLIST(deadftp.queue), g_list_index(ftp_queue, q));
    ftp_queue = g_list_remove(ftp_queue, q);
    queue_item_delete(q);
  }

  /* start the next */
  q = queue_find_next(next);
  if (q) {
    queue_start_item(q);
    return;
  }

  /* There are no more non-error items left */

  /* XXX: Do we want to go back and try to transfer all error 
   * transfers again?  */

  finish_queue();
}

static queue_item_t *queue_find_start()
{
  queue_item_t *q = NULL;

  ftp_queue_current = ftp_queue;
  if (ftp_queue_current)
    q = (queue_item_t *) ftp_queue_current->data;
  while (q && !QUEUE_OK_TRANSFER(q)) {
    ftp_queue_current = g_list_next(ftp_queue_current);
    if (ftp_queue_current)
      q = (queue_item_t *) ftp_queue_current->data;
    else
      q = NULL;
  }
  return q;
}

queue_item_t *queue_find_next(const GList * queue)
{
  queue_item_t *q = NULL;
  GList *queuev = (GList *) queue;

  if (!queue)
    return NULL;

  q = (queue_item_t *) queuev->data;
  while (q && !QUEUE_OK_TRANSFER(q)) {
    queuev = g_list_next(queuev);
    if (queuev)
      q = (queue_item_t *) queuev->data;
    else
      q = NULL;
  }
  if (!q) {			/*
				 * if we don't find one starting with
				 * current, start from beginning of the queue 
				 */
    q = queue_find_start();
    return q;
  }
  ftp_queue_current = queuev;
  return q;
}

void finish_queue()
{
  /* clean up the view */
  queue_hide_status();

  /* Handle to finish_action */
  switch (deadftp_prefs.finish_action) {
  case FINISH_IDLE:
    /* Refresh both local and remote file lists */
    change_local();
    ftp_send_command(FTP_STATE_START_LIST, 0, _("Refreshing Directory List"), "PWD");
    break;
  case FINISH_DISCONNECT:
    /* refresh local file list and disconnect */
    change_local();
    ftp_disconnect();
    break;
  case FINISH_EXIT:
    ftp_exit(0);
    break;
  }
}

queueid_t add_queue(const ftp_server_t * server, struct afile *file,
		    const char *local_dir, const char *dest_file, gboolean send)
{
  return insert_queue(server, file, local_dir, dest_file, send, -1);
}

queueid_t insert_queue(const ftp_server_t * server, struct afile * file,
		       const char *local_dir, const char *dest_file, gboolean send,
		       int position)
{
  queue_item_t *q;
  char buffer[REGULAR_BUFFER];

  g_return_val_if_fail(server && file && server->hostname != NULL, 0);

  q = g_new0(queue_item_t, 1);
  strncpy(q->hostname, server->hostname, SMALL_BUFFER);
  if (server->username && *server->username) {
    strncpy(q->username, server->username, SMALL_BUFFER);
    strncpy(q->password, server->password, SMALL_BUFFER);
  } else {
    strncpy(q->username, "anonymous", SMALL_BUFFER);
    strncpy(q->password, deadftp_prefs.anonymous_password, SMALL_BUFFER);
  }
  q->port = server->port;

  if (send) {
    QUEUE_SET_FLAG(q, QUEUE_FLAG_SEND);
    strncpy(q->local_file, file->file, SMALL_BUFFER);
    strncpy(q->remote_file, dest_file, SMALL_BUFFER);
  } else {
    strncpy(q->local_file, dest_file, SMALL_BUFFER);
    strncpy(q->remote_file, file->file, SMALL_BUFFER);
  }
  strncpy(q->remote_dir, server->current_directory, SMALL_BUFFER);
  strncpy(q->local_dir, local_dir, SMALL_BUFFER);
  if (file->is_dir)
    QUEUE_SET_FLAG(q, QUEUE_FLAG_DIRECTORY);

  q->size = atoi(file->length);
  q->total_trans = 0;
  QUEUE_SET_STATUS(q, QUEUE_STATUS_NORMAL);
  q->mode = str_to_mode(file->mode);

  if (!send && file->date) {
    q->remote_modified = queue_parse_date(file->date);
  } else if (file->date) {
    /* TODO: get modified time for remote file when uploading somehow... */
    q->remote_modified = queue_parse_date(file->date);
  }

  g_snprintf(buffer, REGULAR_BUFFER, "%s/%s", local_dir, file->file);
  QUEUE_CLEAR_FLAG(q, QUEUE_FLAG_RESUME);

  return insert_queue_item(q, position);
}

queueid_t add_queue_item(queue_item_t * q)
{
  return insert_queue_item(q, -1);
}

queueid_t insert_queue_item(queue_item_t * q, int position)
{
  char *filename, *home;
  g_return_val_if_fail(q != NULL, 0);

  home = g_get_home_dir();
  filename = g_strdup_printf("%s/.dead/queue-%d", home, getpid());
  /* We could probably do this a more efficient way since queue_do_save reconstructs
   * the xml DOM structure each time but for now... */
  /* when queue gets big this takes FOREVER, need more effiient way of doing this */
  /* queue_do_save(filename); */
  chmod(filename, S_IRUSR | S_IWUSR);
  g_free(filename);

  /* FIXME: The position code here may be one off, have to check it out */
  ftp_queue = g_list_insert(ftp_queue, (gpointer) q, position);

  if (QUEUE_GET_FLAG(q, QUEUE_FLAG_SEND))
    FTP.queue.upsize += q->size;
  else
    FTP.queue.downsize += q->size;

  q->total_trans = 0;
  q->queueid = queueid++;

  if (IS_SET(FTP_FLAG_FREEZE_QUEUE)
      && !GTK_CLIST(deadftp.queue)->freeze_count) {
    gtk_clist_freeze(GTK_CLIST(deadftp.queue));
  }

  queue_clist_insert(q, position);

  return q->queueid;
}

void queue_error(const char *message)
{
  queue_item_t *q;

  g_return_if_fail(message != NULL);

  if (!ftp_queue_current)
    return;

  q = (queue_item_t *) ftp_queue_current->data;
  QUEUE_SET_STATUS(q, QUEUE_STATUS_ERROR);
  queue_update_item_status(q);

  print_console(CONSOLE_ERROR, "Couldn't transfer file: %s\n", message);

  queue_next(FALSE);
}

void queue_abort()
{
  queue_item_t *q;
  /* Set the current value to the previous so we come to this item again when we start
   * the transfer again
   */
  q = (queue_item_t *) ftp_queue_current->data;
  QUEUE_SET_STATUS(q, QUEUE_STATUS_NORMAL);
  queue_update_item_status(q);

  /* used to calculate the next file to tranfer so set it back so the 
   * current files gets started whe the transfer begins again */
  ftp_queue_current = g_list_previous(ftp_queue_current);
}

queueid_t queue_add_fxp(fxp_desc_t * fxp, const char *hostname, int port,
			const char *username, const char *password,
			const char *remote_dir, const char *remote_file,
			unsigned long size, queueid_t qid)
{
  queue_item_t *q;

  g_return_val_if_fail(hostname != NULL, 0);

  q = g_new0(queue_item_t, 1);
  q->fxp = fxp;
  strncpy(q->hostname, hostname, SMALL_BUFFER);
  q->port = port;
  strncpy(q->username, username, SMALL_BUFFER);
  strncpy(q->password, password, SMALL_BUFFER);
  strncpy(q->remote_dir, remote_dir, SMALL_BUFFER);
  strncpy(q->remote_file, remote_file, SMALL_BUFFER);
  strncpy(q->local_file, remote_file, SMALL_BUFFER);
  q->size = size;
  if (qid) {
    QUEUE_SET_FLAG(q, QUEUE_FLAG_MASTER_ID);
    q->queueid = qid;
  } else
    q->queueid = queueid++;

  ftp_queue = g_list_append(ftp_queue, (gpointer) q);
  if (QUEUE_GET_FLAG(q, QUEUE_FLAG_SEND))
    FTP.queue.upsize += q->size;
  else
    FTP.queue.downsize += q->size;

  q->total_trans = 0;
  queue_clist_append(q);

  return q->queueid;
}

void queue_go_id(gboolean slave, queueid_t qid)
{
  queue_item_t *q = NULL;
  GList *l;

  for (l = ftp_queue; l; l = g_list_next(l)) {
    q = (queue_item_t *) l->data;
    if (((slave && QUEUE_GET_FLAG(q, QUEUE_FLAG_MASTER_ID))
	 || (!slave && !QUEUE_GET_FLAG(q, QUEUE_FLAG_MASTER_ID)))
	&& q->queueid == qid)
      break;
  }
  if (!q)
    return;

  queue_start_item(q);
}

gboolean queue_check_fxp_go()
{
  queue_item_t *next, *cur;

  next = (queue_item_t *) ftp_queue_current->next->data;
  cur = (queue_item_t *) ftp_queue_current->data;
  if (!next)
    return FALSE;
  if (cur->fxp && cur->fxp == next->fxp && cur->fxp->slave)
    return FALSE;
  return TRUE;
}

void queue_item_delete(queue_item_t * q)
{
  if (QUEUE_GET_FLAG(q, QUEUE_FLAG_SEND))
    FTP.queue.upsize -= q->size;
  else
    FTP.queue.downsize -= q->size;

  if (q->fxp)
    fxp_delete(q);
  g_free(q);
}

void queue_item_delete_id(gboolean slave, queueid_t qid)
{
  GList *l;
  queue_item_t *q = NULL;

  for (l = ftp_queue; l; l = g_list_next(l)) {
    q = (queue_item_t *) l->data;
    if (((slave && QUEUE_GET_FLAG(q, QUEUE_FLAG_MASTER_ID))
	 || (!slave && !QUEUE_GET_FLAG(q, QUEUE_FLAG_MASTER_ID)))
	&& q->queueid == qid)
      break;
  }
  if (!q) {
    g_warning("Couldn't find queue item to delete\n");
    return;
  }
  queue_item_delete(q);
}

void queue_calculate_size()
{
  GList *l;
  queue_item_t *q;
  unsigned long usize = 0, dsize = 0;

  for (l = ftp_queue; l; l = g_list_next(l)) {
    q = (queue_item_t *) l->data;
    /* only include files that can/will be transfered in the count */
    if (QUEUE_OK_TRANSFER(q)) {
      if (QUEUE_GET_FLAG(q, QUEUE_FLAG_SEND))
	usize += q->size;
      else
	dsize += q->size;
    }
  }
  FTP.queue.upsize = usize;
  FTP.queue.downsize = dsize;
}

void queue_add_url(const char *url)
{
  char cwd[SMALL_BUFFER], proto[10], file[SMALL_BUFFER], *c;
  struct afile afile;
  ftp_server_t server;
  char *curl = g_strdup(url);

  for (c = strtok(curl, "\n\r"); c; c = strtok(NULL, "\n\r")) {
    if (!parse_url
	(c, proto, 10, server.username, SMALL_BUFFER, server.password, SMALL_BUFFER,
	 server.hostname, SMALL_BUFFER, &server.port, server.current_directory,
	 SMALL_BUFFER, file, SMALL_BUFFER)) {
      return;
    }

    bzero((char *) &afile, sizeof(afile));
    afile.file = file;
    if (g_strcasecmp(proto, "file") == 0) {
      if (FTP.state == FTP_STATE_DISCONNECTED)
	return;
      add_queue(&FTP.server, &afile, server.current_directory, file, TRUE);
    } else if (g_strcasecmp(proto, "ftp") == 0) {
      getcwd(cwd, SMALL_BUFFER);
      add_queue(&server, &afile, cwd, file, FALSE);
    }
  }
}

static gboolean queue_expand_directory(const queue_item_t * q)
{
  GList *l, *pos;
  struct afile *file;
  char cwd[REGULAR_BUFFER];

  pos = g_list_find(ftp_queue, (gpointer) q);
  getcwd(cwd, REGULAR_BUFFER);
  if (QUEUE_GET_FLAG(q, QUEUE_FLAG_SEND))
    l = local_files;
  else
    l = remote_files;

  for (; l; l = g_list_next(l)) {
    file = (struct afile *) l->data;
    if (strncmp(file->file, ".", 1) != 0 && strncmp(file->file, "..", 2) != 0)
      insert_queue(&FTP.server, file, cwd, file->file,
		   QUEUE_GET_FLAG(q, QUEUE_FLAG_SEND), g_list_position(ftp_queue,
								       pos));
  }
  return TRUE;
}

static int queue_parse_month(const char *month)
{
  if (g_strcasecmp(month, "Jan") == 0)
    return 1;
  else if (g_strcasecmp(month, "Feb") == 0)
    return 2;
  else if (g_strcasecmp(month, "Mar") == 0)
    return 3;
  else if (g_strcasecmp(month, "Apr") == 0)
    return 4;
  else if (g_strcasecmp(month, "May") == 0)
    return 5;
  else if (g_strcasecmp(month, "Jun") == 0)
    return 6;
  else if (g_strcasecmp(month, "Jul") == 0)
    return 7;
  else if (g_strcasecmp(month, "Aug") == 0)
    return 8;
  else if (g_strcasecmp(month, "Sep") == 0)
    return 9;
  else if (g_strcasecmp(month, "Oct") == 0)
    return 10;
  else if (g_strcasecmp(month, "Nov") == 0)
    return 11;
  else if (g_strcasecmp(month, "Dec") == 0)
    return 12;
  return 0;
}

static time_t queue_parse_date(const char *date)
{
  time_t t = time(NULL);
  struct tm tim, *timp = NULL;
  char month[4];
  int day, hour, min, year;

  memset((char *) &tim, sizeof(tim), 0);
  if (sscanf(date, "%3s %d %d:%d", month, &day, &hour, &min) == 4) {
    tim.tm_mon = queue_parse_month(month);
    tim.tm_mday = day;
    tim.tm_hour = hour;
    tim.tm_min = min;

    timp = localtime(&t);
    tim.tm_year = timp->tm_year;
  } else if (sscanf(date, "%3s %d %d", month, &day, &year) == 3) {
    tim.tm_mon = queue_parse_month(month);
    tim.tm_mday = day;
    tim.tm_year = year;
  }
  t = mktime(&tim);

  return t;
}
